from coopmovers.segmentfollower import SegmentFollower
from coopstructs.vectors import Vector2
from typing import List
import time
from coopgame.models.status import Status
from coopgame.models.primitives.gameObject import GameObject
from coopgame.models.primitives.health import Health

class Character(GameObject):
    def __init__(self,
                 name,
                 pos,
                 health: Health = None,
                 tags: List[str] = None,
                 max_speed_meters_per_sec: int = 10,
                 max_accel_meters_per_sec: int = 10, initial_velocity: Vector2 = None,
                 statuss: List[Status] = None):
        super().__init__(name=name, pos=pos, tags=tags)
        self.segment_follower = SegmentFollower(name=name
                                                 , start_pos=pos
                                                 , max_speed_meters_per_sec=max_speed_meters_per_sec
                                                 , max_accel_meters_per_sec=max_accel_meters_per_sec
                                                 , initial_velocity=initial_velocity)
        self._health = health
        self._statuss: List[Status] = statuss if statuss else []
        self.name = name

    @property
    def status(self):
        return self._statuss




from coopgame.colors import Color
import pygame
from coopgraph.graphs import Graph, Edge
from coopstructs.geometry.vectors.vectorN import Vector2
from coopstructs.geometry import Rectangle
import coopgame.pygamehelpers as help
from typing import Callable, Tuple, Dict, Iterable
from coopstructs.geometry.curves.curves import LineCurve, Curve
from cooptools.coopEnum import Directionality
from dataclasses import dataclass, asdict
import numpy as np
from coopgame.pointdrawing import point_draw_utils as putils
from coopgame.label_drawing import label_drawing_utils as lutils
from coopgame.linedrawing import line_draw_utils as nutils
from coopstructs.geometry.lines import Line
import cooptools.matrixManipulation as mm
import cooptools.geometry_utils.vector_utils as vec

@dataclass(frozen=True)
class GraphDrawArgs:
    coordinate_converter: Callable[[Vector2], Vector2] = None,
    node_draw_args: putils.DrawPointArgs = None
    enabled_edge_args: nutils.DrawLineArgs = None
    disabled_edge_args: nutils.DrawLineArgs = None
    node_label_args: lutils.DrawLabelArgs = None
    articulation_points_args: putils.DrawPointArgs = None
    source_node_args: putils.DrawPointArgs = None
    sink_node_args: putils.DrawPointArgs = None
    orphan_node_args: putils.DrawPointArgs = None

    def with_(self, **kwargs):
        definition = self.__dict__
        for kwarg, val in kwargs.items():
            definition[kwarg] = val

        return type(self)(**definition)

    def converted_coordinate(self, coord: Tuple[float, float]):
        if self.coordinate_converter is None:
            return coord

        else:
            return self.coordinate_converter(Vector2.from_tuple(coord)).as_tuple()

    @classmethod
    def from_(cls, args, **kwargs):
        kw = asdict(args)
        kw.update(kwargs)
        return GraphDrawArgs(**kw)

    @property
    def EdgesBaseArgs(self):
        return GraphDrawArgs(
            coordinate_converter=self.coordinate_converter,
            enabled_edge_args=self.enabled_edge_args.BaseArgs,
            disabled_edge_args=self.disabled_edge_args.BaseArgs
        )

    @property
    def EdgesLabelArgs(self):
        return GraphDrawArgs(
            coordinate_converter=self.coordinate_converter,
            enabled_edge_args=self.enabled_edge_args.LabelArgs,
            disabled_edge_args=self.disabled_edge_args.LabelArgs
        )

    @property
    def NodesBaseArgs(self):
        return GraphDrawArgs(
            coordinate_converter=self.coordinate_converter,
            node_draw_args=self.node_draw_args
        )

    @property
    def NodesLabelArgs(self):
        return GraphDrawArgs(
            coordinate_converter=self.coordinate_converter,
            node_label_args=self.node_label_args
        )

    @property
    def OverlayArgs(self):
        return GraphDrawArgs(
            coordinate_converter=self.coordinate_converter,
            articulation_points_args=self.articulation_points_args,
            source_node_args=self.source_node_args,
            sink_node_args=self.sink_node_args,
            orphan_node_args=self.orphan_node_args
        )

def draw_to_surface(surface: pygame.Surface,
                    graph: Graph,
                    args: GraphDrawArgs,
                    vec_transformer: vec.VecTransformer = None):
    if graph is None:
        return

    if args.enabled_edge_args or args.disabled_edge_args:
        _draw_edges(
            surface=surface,
            edge_draw_args={
                e: args.enabled_edge_args if e.enabled() else args.disabled_edge_args for e in graph.edges
            },
            vec_transformer=vec_transformer
        )

    if args.node_draw_args or args.node_label_args:
        draw_graph_nodes(surface,
                         graph=graph,
                         node_draw_args=args.node_draw_args,
                         coordinate_converter=args.coordinate_converter,
                         vec_transformer=vec_transformer,
                         draw_label_args=args.node_label_args)

    if args.articulation_points_args:
        putils.draw_points(
            points={args.converted_coordinate(x.pos): args.articulation_points_args for x in graph.AP()},
            surface=surface,
            vec_transformer=vec_transformer
        )

    if args.source_node_args:
        putils.draw_points(
            points={args.converted_coordinate(x.pos): args.source_node_args for x in graph.Sources},
            surface=surface,
            vec_transformer=vec_transformer
        )

    if args.sink_node_args:
        putils.draw_points(
            points={args.converted_coordinate(x.pos): args.sink_node_args for x in graph.Sinks},
            surface=surface,
            vec_transformer=vec_transformer
        )

    if args.orphan_node_args:
        putils.draw_points(
            points={args.converted_coordinate(x.pos): args.orphan_node_args for x in graph.Orphans},
            surface=surface,
            vec_transformer=vec_transformer
        )
def _draw_edges(
    surface: pygame.Surface,
    edge_draw_args: Dict[Edge, nutils.DrawLineArgs],
    vec_transformer: vec.VecTransformer = None
):
    line_args = {Line(Vector2.from_tuple(k.start.pos),
                      Vector2.from_tuple(k.end.pos)): v.with_(label_text=str(k))
                 for k, v in edge_draw_args.items() if k.start.pos != k.end.pos}

    nutils.draw_lines(lines=line_args,
                      surface=surface,
                      vec_transformer=vec_transformer
                      )

def draw_graph_nodes(surface: pygame.Surface,
                     graph: Graph,
                     coordinate_converter: Callable[[Vector2], Vector2] = None,
                     node_draw_args: putils.DrawPointArgs = None,
                     draw_label_args: lutils.DrawLabelArgs = None,
                     vec_transformer: vec.VecTransformer = None):
    if not coordinate_converter:
        coordinate_converter = lambda x: x

    txt_dict = {}
    for node in graph.nodes:
        converted_point = coordinate_converter(Vector2.from_tuple(node.pos)).as_tuple()
        if node_draw_args:
            putils.draw_points({converted_point: node_draw_args},
                               surface=surface,
                               vec_transformer=vec_transformer)


        scaled = vec.resolve_vec_transformed_points([converted_point], vec_transformer=vec_transformer)[0]
        txt_dict.setdefault(scaled, []).append(node.name)


    if draw_label_args:
        for pos, labels in txt_dict.items():
            lutils.draw_label(
                hud=surface,
                text=",".join(labels),
                args=draw_label_args,
                pos=pos
            )


def draw_articulation_points(surface: pygame.Surface,
                             graph: Graph,
                             color: Color = None,
                             vec_transformer: vec.VecTransformer = None
                             ):
    if graph is None:
        return

    articulation_points = graph.AP()

    if color is None:
        color = Color.ORANGE


    scaled_pos = vec.resolve_vec_transformed_points(points=[node.pos for node in articulation_points.keys()],
                                                    vec_transformer=vec_transformer)

    for point in scaled_pos:
        pygame.draw.circle(surface, color.value, (int(point[0]), int(point[1])), 10, 1)

def draw_directionality_indicators(curves: Dict[Curve, Directionality],
                                   surface: pygame.Surface,
                                   indicator_color: Color,
                                   num_arrows: int = 5,
                                   size: float = 1,
                                   indicator_points_color: Color | Tuple[Color, ...]=None,
                                   vec_transformer: vec.VecTransformer = None):
    arrow_ts = [1.0 / (num_arrows - 1) * x for x in range(0, num_arrows)] if num_arrows > 1 else [0.5]


    for curve, direction in curves.items():
        for t in arrow_ts:
            centre = curve.point_at_t(t)

            try:
                # get derivative of curve for drawing
                derivative = curve.derivative_at_t(t)
                d_unit = derivative.unit()
            except:
                # most likely a vertical curve (no derivative), handle by pointing up or down.
                d_unit = (curve.EndPoint - curve.origin).unit()

            if d_unit is None or d_unit.y == 0:
                continue

            d_foreward = d_unit.scaled_to_length(size)
            d_ort_1 = Vector2(1, - d_unit.x / d_unit.y).scaled_to_length(size / 2)
            # d_ort_2 = d_ort_1 * -1

            a = b = c = d = e = f = None
            if direction in [Directionality.FOREWARD, Directionality.BIDIRECTIONAL]:
                a = centre
                b = centre - d_foreward + d_ort_1
                c = centre - d_foreward - d_ort_1

                scaled_polygon_points = vec.resolve_vec_transformed_points(points=[a.as_tuple(), b.as_tuple(), c.as_tuple()],
                                                                           vec_transformer=vec_transformer)

                help.draw_polygon(surface, scaled_polygon_points, color=indicator_color)

            if direction in [Directionality.BACKWARD, Directionality.BIDIRECTIONAL]:
                d = centre
                e = centre + d_foreward + d_ort_1
                f = centre + d_foreward - d_ort_1

                scaled_polygon_points = vec.resolve_vec_transformed_points(points=[d.as_tuple(), e.as_tuple(), f.as_tuple()],
                                                                           vec_transformer=vec_transformer)

                help.draw_polygon(surface, scaled_polygon_points, color=indicator_color)

            ip_color_getter = lambda ii: indicator_points_color[ii % len(indicator_points_color)] \
                if type(indicator_points_color) == tuple \
                else indicator_points_color

            if indicator_points_color:
                points = [a, b, c, d, e, f]
                putils.draw_points(
                    points={
                        x.as_tuple(): putils.DrawPointArgs(
                            color=ip_color_getter,
                            outline_color=None,
                            radius=2
                        ) for ii, x in enumerate(points)
                    },
                    surface=surface,
                    vec_transformer=vec_transformer
                )


from coopgraph.grids import RectGrid
from typing import Hashable, Callable
from coopstructs.curves import LineCurve, Arc, Orientation
from coopstructs.geometry import Rectangle
from coopstructs.geometry.vectors.vectorN import Vector2

class UnitPathHandler():

    def __init__(self, toggled_key: Hashable):
        self.toggled_key = toggled_key

    def generate(self
             , grid: RectGrid
             , grid_box_rect: Rectangle
             , naming_provider: Callable[[], str]):

        curves = []

        for row in range(0, grid.nRows):
            for col in range(0, grid.nColumns):

                cell = grid.at(row, col)
                if not cell.state[self.toggled_key].value:
                    continue

                left = cell.left.state[self.toggled_key].value if cell.left else None
                right = cell.right.state[self.toggled_key].value if cell.right else None
                up = cell.up.state[self.toggled_key].value if cell.up else None
                down = cell.down.state[self.toggled_key].value if cell.down else None
                connection_count = sum(x for x in [left, right, up, down] if x)

                if left and right:
                    # Horizontal Line
                    start = Vector2(col * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                    end = Vector2((col + 1) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                if up and down:

                    # Vertical Line
                    start = Vector2((col + .5) * grid_box_rect.width, row * grid_box_rect.height)
                    end = Vector2((col + 0.5) * grid_box_rect.width, ((row + 1) * grid_box_rect.height))

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                if up and right:
                    origin = Vector2((col + .5) * grid_box_rect.width, (row) * grid_box_rect.height)
                    curves.append(Arc(id = naming_provider(),
                                      orientation=Orientation.DOWN_RIGHT,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))

                if up and left:
                    origin = Vector2((col) * grid_box_rect.width, (row + 0.5) * grid_box_rect.height)
                    curves.append(Arc(id=naming_provider(),
                                      orientation=Orientation.RIGHT_UP,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))

                if right and down:
                    origin = Vector2((col + .5) * grid_box_rect.width, (row + 1) * grid_box_rect.height)
                    curves.append(Arc(id=naming_provider(),
                                      orientation=Orientation.UP_RIGHT,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))


                if down and left:
                    origin = Vector2((col) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                    curves.append(Arc(id=naming_provider(),
                                      orientation=Orientation.RIGHT_DOWN,
                                      origin=origin,
                                      arc_box_size=Vector2(grid_box_rect.width, grid_box_rect.height)))

                if connection_count == 1:
                    if up:
                        # Vertical Line
                        start = Vector2((col + .5) * grid_box_rect.width, row * grid_box_rect.height)
                        end = Vector2((col + 0.5) * grid_box_rect.width, ((row + 0.5) * grid_box_rect.height))

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))
                    elif down:
                        # Vertical Line
                        start = Vector2((col + .5) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                        end = Vector2((col + 0.5) * grid_box_rect.width, ((row + 1) * grid_box_rect.height))

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                    elif right:
                        # Horizontal Line
                        start = Vector2((col + .5) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                        end = Vector2((col + 1) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                    elif left:
                        # Horizontal Line
                        start = Vector2(col * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                        end = Vector2((col + .5) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                        curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                elif connection_count == 0:
                    # Horizontal Line
                    start = Vector2((col + 0.25) * grid_box_rect.width, (row + .5) * grid_box_rect.height)
                    end = Vector2((col + .75) * grid_box_rect.width, (row + .5) * grid_box_rect.height)

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

                    # Vertical Line
                    start = Vector2((col + .5) * grid_box_rect.width, (row + .25) * grid_box_rect.height)
                    end = Vector2((col + .5) * grid_box_rect.width, (row + .75) * grid_box_rect.height)

                    curves.append(LineCurve(id=naming_provider(), origin=start, destination=end))

        return curves


from cooptools.register import Register
from typing import Protocol, Dict

class ComponentProtocol(Protocol):
    def on_update(self, delta_time: int, **kwargs):
        pass

class EntityComponentSystem:

    def __init__(self, components: Dict[str, ComponentProtocol] = None):
        self._component_registry = Register[ComponentProtocol]()

        if components is not None:
            self.register_components(components)

    def register_components(self, components: Dict[str, ComponentProtocol]):
        self._component_registry.register(list(components.values()), ids=list(components.keys()))

    def update(self, delta_time: int, **kwargs):
        for id, component in self._component_registry.Registry.items():
            component.on_update(delta_time, **kwargs)

    @property
    def Components(self) -> Dict[str, ComponentProtocol]:
        return self._component_registry.Registry

if __name__ == "__main__":
    class Health:
        def __init__(self):
            self.health = 0

        def on_update(self, delta_time, amount):
            self.health += amount

        def __repr__(self):
            return str(self.health)

    ecs = EntityComponentSystem({'health': Health()})

    print(ecs.Components)

    ecs.update(500, amount=5)

    print(ecs.Components)




import functools
import logging
from coopbugger.monitoredclass import MonitoredClass
import time
from typing import Callable, Tuple
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle
from coopgame.colors import Color
from coopgame.runtime_config import RuntimeConfig
# from coopgame.inputListener import InputState

def try_handler(func):
    @functools.wraps(func)
    def wrapper_handler(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except NotImplementedError as e:
            error = f"Inherited class should implement logic for {func.__name__}"
            logging.warning(error)
        except Exception as e:
            logging.exception(e)

    return wrapper_handler


class IGameTemplate(MonitoredClass):

    def __init__(self,
                 config: RuntimeConfig = None):
        super().__init__()
        self.runtime_config = config if config else RuntimeConfig()

        self.screen = None
        self._init_screen(self.runtime_config)

        self.ticks = 0
        self.frame_times = []
        self.fps = None

        self.buttons = {}
        self._key_handlers = {}
        self.register_keys()

        self.running = False

        self.input_state = InputState()

        self.time_since_logged_tracked_time = 0

    @property
    def screen_width(self):
        return self.runtime_config.screen_dims[0]

    @property
    def screen_height(self):
        return self.runtime_config.screen_dims[1]

    def _break_run_loop(self):
        self.running = False

    def toggle_fullscreen(self):
        self.fullscreen = not self.fullscreen
        self._init_screen(self.fullscreen)


    def main(self):

        self.initialize_game()

        self.running = True
        ii = 0


        while self.running:
            ''' Calculate the ticks between update calls so that the update functions can handle correct time deltas '''
            t = time.perf_counter() * 1000
            delta_time_ms = int(t - self.ticks)
            self.ticks = t

            self._update(delta_time_ms)
            self._draw(frames=ii)
            self._log_tracked_time(delta_time_ms)
            ii += 1

        self._quit()

    def calculate_fps(self, ticks_last_frame: int):
        if len(self.frame_times) > 20:
            self.frame_times.pop(0)

        self.frame_times.append(ticks_last_frame)

        avg_sec_per_frame = sum(self.frame_times) / len(self.frame_times) / 1000.0
        self.fps = 1 / avg_sec_per_frame if avg_sec_per_frame > 0 else 0

    def _update(self, delta_time_ms: int):
        """:return
            Update environment based on time delta and any input
        """
        self.calculate_fps(delta_time_ms)

        '''Update Model'''
        self.model_updater(delta_time_ms)

        '''Update Sprites'''
        self.sprite_updater(delta_time_ms)

        '''Handle Events'''
        self._handle_events()

        '''Clear Input Events'''
        self.input_state.clear_events()

    def _handle_events(self):
        self._handle_library_specific_events()
        self._handle_key_pressed()
        self._handle_held_keys()
        self._handle_hover_over()

    def _log_tracked_time(self, delta_time_ms: int):
        self.time_since_logged_tracked_time += delta_time_ms / 1000.0

        if self.time_since_logged_tracked_time > self.runtime_config.log_tracked_time_interval_sec:
            logging.info(self.tracked_time)
            self.time_since_logged_tracked_time = 0

    def register_action_to_keys(self, keys_tuple, func: Callable, react_while_holding: bool = False):
        """
            Takes a tuple of keys (integers) and maps that key combo to a callable function

            :param keys_tuple a list of keys (integers) that will be mapped to the input callable. Note that a single
            value Tuple is input as ([key],) *note the comma
            :param func a callable that is mapped to the key combo
        """
        self._key_handlers[keys_tuple] = (func, react_while_holding)

    @MonitoredClass.timer
    @try_handler
    def handle_buttons(self, event):
        for id, button in self.buttons:
            if 'click' in button.handleEvent(event):
                button.callback()

    @MonitoredClass.timer
    @try_handler
    def _handle_key_pressed(self):
        for key in self.input_state.keys_pressed_events:
            logging.info(f"{key} pressed")

        for mapped_keys, reaction in self._key_handlers.items():
            if self.input_state.is_pressed(mapped_keys):
                func = reaction[0]
                logging.info(f"callback triggered for [{mapped_keys}] --> {func}")
                func()

    @MonitoredClass.timer
    @try_handler
    def _handle_held_keys(self):
        for mapped_keys, reaction in self._key_handlers.items():
            if self.input_state.is_pressed(mapped_keys):
                func = reaction[0]
                react_while_holding = reaction[1]
                if react_while_holding:
                    logging.info(f"callback triggered for [{mapped_keys}] (while held)--> {func}")
                    func()

    @MonitoredClass.timer
    def _handle_hover_over(self):
        mouse_pos_as_vector = self.mouse_pos_as_vector()
        if mouse_pos_as_vector:
            self.handle_hover_over(mouse_pos_as_vector)

    @MonitoredClass.timer
    def _draw(self,
              frames: int):
        self._init_frame()
        self.draw(frames)
        if self.runtime_config.debug_mode:
            self._draw_debug_stuff()

        self._render_frame()

    @MonitoredClass.timer
    def _draw_debug_stuff(self):
        # Monitored class stats
        self._draw_monitoredclass_stats(self, self.runtime_config)

        # FPS
        self._draw_fps(self.runtime_config)

        # Mouse Coords
        self.draw_mouse_coord(self.runtime_config)

    @try_handler
    def draw_mouse_coord(self,
                         config: RuntimeConfig):
        mouse_pos_as_vector = self.mouse_pos_as_vector()

        offset_rect = Rectangle(self.runtime_config.mousecoords_txt_offset_x,
                                self.runtime_config.mousecoords_txt_offset_y,
                                1, 1)

        txt = f"M:<{int(mouse_pos_as_vector.x)}, {int(mouse_pos_as_vector.y)}>"
        self._draw_text(txt,
                        offset_rect=offset_rect,
                        color=config.mousecoords_txt_color,
                        font_size=config.mousecoords_txt_font_size
                        )

    def _draw_fps(self, config: RuntimeConfig):
        fps_offset_rect = Rectangle(self.runtime_config.fps_txt_offset_x,
                                    self.runtime_config.fps_txt_offset_y,
                                    1, 1)
        txt = f"FPS: {int(self.fps)}"
        self._draw_text(txt,
                        offset_rect=fps_offset_rect,
                        color=config.fps_color,
                        font_size=config.fps_txt_font_size)

    def window_rect(self):
        return Rectangle(0, 0, self.screen.get_height(), self.screen.get_width())

    def toggle_debug_mode(self):
        self._debug_mode = not self._debug_mode


    ############### IGAMETEMPLATE IMPLEMNENTER SPECIFIC LOGIC ###############

    @MonitoredClass.timer
    @try_handler
    def initialize_game(self):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_left_click(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_right_click(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_mouse_scroll_up(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def handle_mouse_scroll_down(self, pressed_keys):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def draw(self, frames: int, debug_mode: bool = False):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def model_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def sprite_updater(self, delta_time_ms: int):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def on_resize(self):
        raise NotImplementedError()

    @try_handler
    def register_keys(self):
        raise NotImplementedError()

    def _quit(self):
        raise NotImplementedError()

    def _init_screen(self, fullscreen):
        raise NotImplementedError()

    def register_button(self, id, text, callback, postion_rect):
        raise NotImplementedError()

    def _handle_library_specific_events(self):
        raise NotImplementedError()


    @MonitoredClass.timer
    @try_handler
    def _init_frame(self):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def _render_frame(self):
        raise NotImplementedError()

    @MonitoredClass.timer
    @try_handler
    def _draw_text(self,
                  text: str,
                  offset_rect: Rectangle = None,
                  color: Color = None,
                  **kwargs):
        raise NotImplementedError()

    @try_handler
    def _draw_monitoredclass_stats(self,
                                  monitoredClass: MonitoredClass,
                                  config: RuntimeConfig):
        raise NotImplementedError()


    @staticmethod
    def mouse_pos_as_vector() -> Vector2:
        raise NotImplementedError()



# This class handles sprite sheets
# This was taken from www.scriptefun.com/transcript-2-using
# sprite-sheets-and-drawing-the-background
# I've added some code to fail if the file wasn't found..
# Note: When calling images_at the rect is the format:
# (x, y, x + offset, y + offset)

# Additional notes
# - Further adaptations from https://www.pygame.org/wiki/Spritesheet
# - Cleaned up overall formatting.
# - Updated from Python 2 -> Python 3.

import pygame
import coopgame.spriteHandling.utils as utils
from typing import List

class SpriteSheet:

    def __init__(self, filename, n_rows, n_columns, x_margin_left=0, x_margin_right=None, x_padding=0,
                         y_margin_top=0, y_margin_bottom = None, y_padding=0):
        self.file_size = utils.get_image_size(filename)
        self.n_images_in_row = n_columns
        self.n_images_in_column = n_rows

        self.x_margin_left = x_margin_left
        self.x_margin_right = x_margin_right if x_margin_right else x_margin_left
        self.x_padding = x_padding
        self.y_margin_top = y_margin_top
        self.y_margin_bottom = y_margin_bottom if y_margin_bottom else y_margin_top
        self.y_padding = y_padding

        self.pixel_width = (self.file_size[0] - (self.x_margin_left + self.x_margin_right)
                         - (self.n_images_in_row - 1) * self.x_padding) // self.n_images_in_row
        self.pixel_height = (self.file_size[1] - (self.y_margin_top + self.y_margin_bottom)
                         - (self.n_images_in_column - 1) * self.y_padding) // self.n_images_in_column


        """Load the sheet."""
        self.sheet = utils.try_load_sprite_file(filepath=filename)

    def load_row_strip(self, row, colorkey = None, n_images: int = None) -> List[pygame.Surface]:
        grid_images = self.load_grid_images(colorkey=colorkey)

        if n_images is None or n_images < 0:
            return grid_images[row]
        else:
            return grid_images[row][:n_images]

    def load_column_strip(self, column, colorkey = None, n_images: int = None):
        grid_images = self.load_grid_images(colorkey=colorkey)

        images =  [row[column] for row in grid_images]
        if n_images is None or n_images < 0:
            return images
        else:
            return images[:n_images]

    def load_grid_images(self, colorkey = None):
        """Load a grid of images.
        x_margin is space between top of sheet and top of first row.
        x_padding is space between rows.
        Assumes symmetrical padding on left and right.
        Same reasoning for y.
        Calls self.images_at() to get list of images.
        """
        sheet_rect = self.sheet.get_rect()
        sheet_width, sheet_height = sheet_rect.size

        # To calculate the size of each sprite, subtract the two margins,
        #   and the padding between each row, then divide by num_cols.
        # Same reasoning for y.
        # x_sprite_size = (sheet_width - 2 * self.x_margin
        #                  - (self.n_images_in_row - 1) * self.x_padding) // self.n_images_in_row
        # y_sprite_size = (sheet_height - 2 * self.y_margin
        #                  - (self.n_images_in_column - 1) * self.y_padding) // self.n_images_in_column

        grid_images = []
        for row_num in range(self.n_images_in_column):
            grid_images.append([])
            for col_num in range(self.n_images_in_row):
                # Position of sprite rect is margin + one sprite size
                #   and one padding size for each row. Same for y.
                x = self.x_margin_left + 1 + col_num * (self.pixel_width + self.x_padding)
                y = self.y_margin_top + 1 + row_num * (self.pixel_height + self.y_padding)
                sprite_rect = x, y, self.pixel_width, self.pixel_height
                grid_images[row_num].append(utils.image_at_rect(surf=self.sheet, rectangle=sprite_rect, colorkey=colorkey))

        return grid_images


from coopgame.particleSystem import Particle
from typing import Callable, Union
from dataclasses import dataclass

particle_destroyer = Callable[[Particle], bool]


@dataclass
class CompareArgs:
    lt: int = None
    gt: int = None
    le: int = None
    ge: int = None

    def compare(self, val):
        if self.lt is not None and val < self.lt:
            return True

        if self.gt is not None and val > self.gt:
            return True

        if self.le is not None and val <= self.le:
            return True

        if self.ge is not None and val >= self.ge:
            return True

        return False


def time_destroyer(particle: Particle, lifespan_ms: Union[int, Callable]) -> bool:
    if callable(lifespan_ms):
        lifespan_ms = lifespan_ms()

    return CompareArgs(gt=lifespan_ms).compare(particle.Lifetime_ms)

def size_destroyer(particle: Particle, compareArgs: CompareArgs):
    return compareArgs.compare(particle.Size)

def time_and_size_destroyer(particle: Particle,
                            lifespan_ms: Union[int, Callable],
                            size_compare_args: CompareArgs) -> bool:
    return time_destroyer(particle=particle, lifespan_ms=lifespan_ms) and \
           size_destroyer(particle=particle, compareArgs=size_compare_args)



from coopgame.colors import Color
import pygame
from coopgraph.graphs import Graph
import coopgame.pygamehelpers as help
from typing import List, Callable
import coopgame.grids_graphs.draw_graph_utils as utils
import coopgame.linedrawing.line_draw_utils as lutils
import coopgame.label_drawing.label_drawing_utils as labutils
from cooptools.coopEnum import CoopEnum
from coopgame.surfaceManager import SurfaceManager, SurfaceRegistryArgs
import coopgame.pointdrawing.point_draw_utils as putils
import cooptools.geometry_utils.vector_utils as vec

class GraphSurfaceType(CoopEnum):
    EDGES_SURFACE_ID = 'EDGES_SURFACE_ID'
    NODES_SURFACE_ID = 'NODES_SURFACE_ID'
    NODE_LABELS_SURFACE_ID = 'NODE_LABELS_SURFACE_ID'
    EDGE_LABELS_SURFACE_ID = 'EDGE_LABELS_SURFACE_ID'
    OVERLAY_SURFACE_ID = 'OVERLAY_SURFACE_ID'

DEFAULT_DRAW_CONFIG = utils.GraphDrawArgs(
            coordinate_converter= lambda x: x,
            node_draw_args=putils.DrawPointArgs(
                color=Color.DARK_BLUE,
                radius=5,
            ),
            enabled_edge_args=lutils.DrawLineArgs(
                color=Color.ORANGE,
                directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                    color=Color.ORANGE,
                    height=10,
                    width=10
                ),
                draw_label_args=labutils.DrawLabelArgs(
                    color=Color.WHEAT
                )
            ),
            disabled_edge_args=lutils.DrawLineArgs(
                color=Color.BROWN,
                directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                    color=Color.BROWN,
                    height=10,
                    width=10
                ),
                draw_label_args=labutils.DrawLabelArgs(
                    color=Color.BROWN
                )
            ),
            node_label_args=labutils.DrawLabelArgs(
                color=Color.WHEAT
            ),
            articulation_points_args=putils.DrawPointArgs(
                outline_color=Color.PURPLE,
                radius=10,
                outline_width=3
            ),
            sink_node_args=putils.DrawPointArgs(
                color=Color.RED,
                radius=10
            ),
            source_node_args=putils.DrawPointArgs(
                color=Color.GREEN,
                radius=10
            ),
            orphan_node_args=putils.DrawPointArgs(
                outline_color=Color.RED,
                radius=10
            )
        )

GraphGetter = Callable[[], Graph]

class PyGraphHandler:
    def __init__(self,
                 screen: pygame.Surface,
                 graph_getter: GraphGetter,
                 draw_config: utils.GraphDrawArgs = None,
                 vec_transformer: vec.VecTransformer = None):
        self.parent_screen = screen
        self.graph_getter = graph_getter
        self._draw_config: utils.GraphDrawArgs = draw_config if draw_config else DEFAULT_DRAW_CONFIG
        self._vec_transformer = vec_transformer

        self.surface_manager = SurfaceManager(
            surface_draw_callbacks=[
                SurfaceRegistryArgs(GraphSurfaceType.EDGES_SURFACE_ID.value, self.redraw_edges_surface),
                SurfaceRegistryArgs(GraphSurfaceType.NODES_SURFACE_ID.value, self.redraw_nodes_surface),
                SurfaceRegistryArgs(GraphSurfaceType.EDGE_LABELS_SURFACE_ID.value, self.redraw_edge_labels_surface),
                SurfaceRegistryArgs(GraphSurfaceType.NODE_LABELS_SURFACE_ID.value, self.redraw_node_labels_surface),
                SurfaceRegistryArgs(GraphSurfaceType.OVERLAY_SURFACE_ID.value, self.redraw_overlay_surface),
            ]
        )

    def set_config(self,
                   draw_config: utils.GraphDrawArgs):
        self._draw_config = draw_config

    def redraw_edges_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.EdgesBaseArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_nodes_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.NodesBaseArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_node_labels_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.NodesLabelArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_edge_labels_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.EdgesLabelArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_overlay_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.OverlayArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def update(self):
        # No surfaces require redrawing every iteration
        pass

    def invalidate(self):
        self.redraw()

    def redraw(self):
        self.surface_manager.redraw([x.name for x in GraphSurfaceType])

    def render(self,
               surface: pygame.Surface):
        self.update()
        self.surface_manager.render(surface)

    def toggle_surface(self, graphSurfaceTypes: List[GraphSurfaceType]):
        self.surface_manager.toggle_visible([x.value for x in graphSurfaceTypes])

    def show_all(self):
        self.surface_manager.show_all()

    def hide_all(self):
        self.surface_manager.hide_all()



from .particle import Particle
from .emitter import Emitter
from .destroyers import *
from .recipies import *
from .utils import *
from coopgame.particleSystem.rendering.colorBehaviors import *
from .rendering import *
from .particleGenerationArgs import ParticleGenerationArgs, BurstArgs



class MissingArgumentException(Exception):
    def __init__(self,
                 argument_name: str,
                 intent: str = None
                 ):
        self.argument_name = argument_name
        self.intent = intent
        super().__init__()

    def __str__(self):
        return f"The named argument {self.argument_name} was missing from the provided **kwargs set. {self.intent if self.intent else ''}"


from enum import Enum
from typing import Tuple
import random as rnd

class Color(Enum):
    # LEGACY
    LIGHT_RED = 255, 204, 204
    BRIGHT_BLUE = 0, 100, 255

    # https: // www.colorschemer.com / rgb - color - codes /
    BLACK = 0, 0, 0
    NAVY = 0, 0, 128
    DARK_BLUE = 0, 0, 139
    MEDIUM_BLUE = 0, 0, 205
    BLUE = 0, 0, 255
    DARK_GREEN = 0, 100, 0
    GREEN = 0, 128, 0
    TEAL = 0, 128, 128
    DARK_CYAN = 0, 139, 139
    DEEP_SKY_BLUE = 0, 191, 255
    DARK_TURQUOISE = 0, 206, 209
    MEDIUM_SPRING_GREEN = 0, 250, 154
    LIME = 0, 255, 0
    SPRING_GREEN = 0, 255, 127
    AQUA = 0, 255, 255
    CYAN = 0, 255, 255
    CORN_FLOWER_BLUE = 100, 149, 237
    MEDIUM_AQUA_MARINE = 102, 205, 170
    DIM_GRAY = 105, 105, 105
    DIM_GREY = 105, 105, 105
    SLATE_BLUE = 106, 90, 205
    OLIVE_DRAB = 107, 142, 35
    SLATE_GRAY = 112, 128, 144
    LIGHT_SLATE_GRAY = 119, 136, 153
    MEDIUM_SLATE_BLUE = 123, 104, 238
    LAWN_GREEN = 124, 252, 0
    CHART_REUSE = 127, 255, 0
    AQUA_MARINE = 127, 255, 212
    MAROON = 128, 0, 0
    PURPLE = 128, 0, 128
    OLIVE = 128, 128, 0
    GRAY = 128, 128, 128
    GREY = 128, 128, 128
    SKY_BLUE = 135, 206, 235
    LIGHT_SKY_BLUE = 135, 206, 250
    BLUE_VIOLET = 138, 43, 226
    DARK_RED = 139, 0, 0
    DARK_MAGENTA = 139, 0, 139
    SADDLE_BROWN = 139, 69, 19
    DARK_SEA_GREEN = 143, 188, 143
    LIGHT_GREEN = 144, 238, 144
    MEDIUM_PURPLE = 147, 112, 219
    DARK_VIOLET = 148, 0, 211
    PALE_GREEN = 152, 251, 152
    DARK_ORCHID = 153, 50, 204
    YELLOW_GREEN = 154, 205, 50
    SIENNA = 160, 82, 45
    BROWN = 165, 42, 42
    DARK_GRAY = 169, 169, 169
    DARK_GREY = 169, 169, 169
    LIGHT_BLUE = 173, 216, 230
    GREEN_YELLOW = 173, 255, 47
    PALE_TURQUOISE = 175, 238, 238
    LIGHT_STEEL_BLUE = 176, 196, 222
    POWDER_BLUE = 176, 224, 230
    FIREBRICK = 178, 34, 34
    DARK_GOLDEN_ROD = 184, 134, 11
    MEDIUM_ORCHID = 186, 85, 211
    ROSY_BROWN = 188, 143, 143
    DARK_KHAKI = 189, 183, 107
    SILVER = 192, 192, 192
    MEDIUM_VIOLET_RED = 199, 21, 133
    PERU = 205, 133, 63
    INDIAN_RED = 205, 92, 92
    CHOCOLATE = 210, 105, 30
    TAN = 210, 180, 140
    LIGHT_GRAY = 211, 211, 211
    LIGHT_GREY = 211, 211, 211
    THISTLE = 216, 191, 216
    ORCHID = 218, 112, 214
    GOLDEN_ROD = 218, 165, 32
    PALE_VIOLET_RED = 219, 112, 147
    CRIMSON = 220, 20, 60
    GAINSBORO = 220, 220, 220
    PLUM = 221, 160, 221
    BURLY_WOOD = 222, 184, 135
    LIGHT_CYAN = 224, 255, 255
    LAVENDER = 230, 230, 250
    DARK_SALMON = 233, 150, 122
    VIOLET = 238, 130, 238
    PALE_GOLDEN_ROD = 238, 232, 170
    LIGHT_CORAL = 240, 128, 128
    KHAKI = 240, 230, 140
    ALICE_BLUE = 240, 248, 255
    HONEYDEW = 240, 255, 240
    AZURE = 240, 255, 255
    SANDY_BROWN = 244, 164, 96
    WHEAT = 245, 222, 179
    BEIGE = 245, 245, 220
    WHITE_SMOKE = 245, 245, 245
    MINT_CREAM = 245, 255, 250
    GHOST_WHITE = 248, 248, 255
    MIDNIGHT_BLUE = 25, 25, 112
    SALMON = 250, 128, 114
    ANTIQUE_WHITE = 250, 235, 215
    LINEN = 250, 240, 230
    LIGHT_GOLDEN_ROD_YELLOW = 250, 250, 210
    OLD_LACE = 253, 245, 230
    RED = 255, 0, 0
    FUCHSIA = 255, 0, 255
    MAGENTA = 255, 0, 255
    HOT_PINK = 255, 105, 180
    CORAL = 255, 127, 80
    DARK_ORANGE = 255, 140, 0
    LIGHT_SALMON = 255, 160, 122
    ORANGE = 255, 165, 0
    LIGHT_PINK = 255, 182, 193
    PINK = 255, 192, 203
    DEEP_PINK = 255, 20, 147
    GOLD = 255, 215, 0
    PEACH_PUFF = 255, 218, 185
    NAVAJO_WHITE = 255, 222, 173
    MOCCASIN = 255, 228, 181
    BISQUE = 255, 228, 196
    MISTY_ROSE = 255, 228, 225
    BLANCHED_ALMOND = 255, 235, 205
    PAPAYA_WHIP = 255, 239, 213
    LAVENDER_BLUSH = 255, 240, 245
    SEA_SHELL = 255, 245, 238
    CORN_SILK = 255, 248, 220
    LEMON_CHIFFON = 255, 250, 205
    FLORAL_WHITE = 255, 250, 240
    SNOW = 255, 250, 250
    YELLOW = 255, 255, 0
    LIGHT_YELLOW = 255, 255, 224
    IVORY = 255, 255, 240
    WHITE = 255, 255, 255
    ORANGE_RED = 255, 69, 0
    TOMATO = 255, 99, 71
    DODGER_BLUE = 30, 144, 255
    LIGHT_SEA_GREEN = 32, 178, 170
    FOREST_GREEN = 34, 139, 34
    SEA_GREEN = 46, 139, 87
    DARK_SLATE_GRAY = 47, 79, 79
    LIME_GREEN = 50, 205, 50
    MEDIUM_SEA_GREEN = 60, 179, 113
    TURQUOISE = 64, 224, 208
    ROYAL_BLUE = 65, 105, 225
    STEEL_BLUE = 70, 130, 180
    MEDIUM_TURQUOISE = 72, 209, 204
    DARK_SLATE_BLUE = 72, 61, 139
    INDIGO = 75, 0, 130
    DARK_OLIVE_GREEN = 85, 107, 47
    CADET_BLUE = 95, 158, 160

    @classmethod
    def closest_color(cls, rgb: Tuple[int, int, int]):
        best_mse = None
        best_color = None
        for color in cls:
            dr = rgb[0] - color.value[0]
            dg = rgb[1] - color.value[1]
            db = rgb[2] - color.value[2]
            mse = dr * dr + dg * dg + db * db

            if mse == 0:
                return color
            elif best_mse is None or mse < best_mse:
                best_mse = mse
                best_color = color

        return best_color

    @classmethod
    def furthest_color(cls, rgb: Tuple[int, int, int]):
        worst_mse = None
        worst_color = None
        for color in cls:
            dr = rgb[0] - color.value[0]
            dg = rgb[1] - color.value[1]
            db = rgb[2] - color.value[2]
            mse = dr * dr + dg * dg + db * db

            if worst_mse is None or mse > worst_mse:
                worst_mse = mse
                worst_color = color

        return worst_color



    @classmethod
    def random(cls):
        r = rnd.randint(0, 255)
        g = rnd.randint(0, 255)
        b = rnd.randint(0, 255)

        return cls.closest_color((r, g, b))



if __name__ == "__main__":
    close = Color.closest_color((45, 45, 45))

    print(f"Color: {close}: {close.value}")


import pygame
from dataclasses import dataclass, asdict
from cooptools.colors import Color
from coopgame.label_drawing.pyLabel import TextAlignmentType
from cooptools.anchor import Anchor2D, CardinalPosition
import logging
from typing import Tuple, Dict
from coopstructs.geometry import Rectangle
import coopgame.pygamehelpers as help
from coopgame.pointdrawing import point_draw_utils as putils

logger = logging.getLogger('coopgame.labeldraw')

@dataclass(frozen=True)
class DrawLabelArgs:
    color: Color = None
    font: pygame.font.Font = None
    alignment: TextAlignmentType = None
    anchor_alignment: CardinalPosition = None
    alpha: float = None
    offset: Tuple[int, int] = None

    def with_(self, **kwargs):
        kw = asdict(self)
        kw.update(kwargs)
        return DrawLabelArgs(**kw)

def draw_label(hud: pygame.Surface,
               text: str,
               args: DrawLabelArgs,
               pos: Tuple[float, float] = None,
               offset_rect: Rectangle = None,
               deb: bool = False
               ):
    font = args.font
    if font is None:
        font = pygame.font.Font(None, 20)

    if pos is None and offset_rect is None:
        raise ValueError(f"At least one of pos and rect cannot be none")

    if offset_rect is None:
        pos = [int(x) for x in pos[:2]]
        if args.offset:
            pos = pos[0] + args.offset[0], pos[1] + args.offset[1]

        anchor_alignment = args.anchor_alignment if args.anchor_alignment is not None else CardinalPosition.TOP_LEFT

        offset_rect = Rectangle(
            anchor=Anchor2D(pt=pos,
                            dims=font.size(text),
                            cardinality=anchor_alignment,
                            inverted_y=True
                            )
        )

    color=args.color
    if color is None:
        color = Color.BLUE

    alignment = args.alignment
    if alignment is None:
        alignment = TextAlignmentType.TOPLEFT

    rendered_txt = font.render(text, True, color.value)

    alpha = args.alpha
    if alpha:
        rendered_txt.set_alpha(alpha * 100)

    if deb:
        help.draw_box(
            surface=hud,
            rect=offset_rect,
            outline_color=Color.PINK,
            anchor_draw_args=putils.DrawPointArgs(
                color=Color.HOT_PINK,
                radius=4
            ),
            corner_draw_args=putils.DrawPointArgs(
                color=Color.LIGHT_BLUE,
                radius=1
            )
        )

    try:
        rect = rendered_txt.get_rect()
        align_coords = alignment_coords_for_type_rect(alignment, offset_rect)
        setattr(rect, alignment.value, align_coords)
        hud.blit(rendered_txt, rect)
        return rendered_txt
    except Exception as e:
        logger.error(f"{e}")

def alignment_coords_for_type_rect(alignment: TextAlignmentType, rect: Rectangle) -> Tuple[float, float]:
    """
    Note the orientation shift from top to bottom bc of the pygame inversion
    """
    switch = {
        TextAlignmentType.TOPRIGHT:      lambda: rect.TopRight,
        TextAlignmentType.TOPLEFT:       lambda: rect.TopLeft,
        TextAlignmentType.TOPCENTER:     lambda: rect.TopCenter,
        TextAlignmentType.BOTTOMLEFT:    lambda: rect.BottomLeft,
        TextAlignmentType.BOTTOMRIGHT:   lambda: rect.BottomRight,
        TextAlignmentType.RIGHTCENTER:   lambda: rect.RightCenter,
        TextAlignmentType.BOTTOMCENTER:  lambda: rect.BottomCenter,
        TextAlignmentType.LEFTCENTER:    lambda: rect.LeftCenter,
        TextAlignmentType.CENTER:        lambda: rect.Center
    }

    return switch.get(alignment)()

def draw_dict(dict_to_draw: Dict,
              draw_args: DrawLabelArgs,
              surface: pygame.Surface,
              total_game_time_sec: float,
              title: str = None,
              inter_line_buffer: int = 3,
              deb: bool=False):

    tracked_time = [(key, val) for key, val in dict_to_draw.items()]
    tracked_time.sort(key=lambda x: x[1], reverse=True)

    txt_lmbda = lambda key, val: f"{key}: {round(val, 2)} sec ({round(val / total_game_time_sec * 100, 1)}%)"

    rendered_txt = draw_label(
        hud=surface,
        text=title,
        args=draw_args,
        pos=(0, 0),
        deb=deb
    )

    y_off = rendered_txt.get_rect().height + inter_line_buffer
    for key, val in tracked_time:

        rendered_txt = draw_label(
            hud=surface,
            text=txt_lmbda(key, val),
            args=draw_args,
            pos=(0, y_off),
            deb=deb
        )

        y_off += rendered_txt.get_rect().height + inter_line_buffer

    return y_off


import pygame
from cooptools.coopEnum import CoopEnum
from coopgame.surfaceManager import SurfaceManager, SurfaceRegistryArgs
from typing import Callable, List, Dict
from coopstructs.geometry.curves.curves import Curve
import coopgame.linedrawing.curve_draw_utils as utils
import coopgame.pygamehelpers as help
import cooptools.geometry_utils.vector_utils as vec

class CurveSurfaceType(CoopEnum):
    STATIC_BUFFER_ID = 'STATIC_BUFFER_ID'
    DYNAMIC_BUFFER_ID = 'DYNAMIC_BUFFER_ID'
    STATIC_CURVES_ID = 'STATIC_CURVES_ID'
    DYNAMIC_CURVES_ID = 'DYNAMIC_CURVES_ID'
    STATIC_OVERLAY_ID = 'OVERLAY_ID'
    DYNAMIC_OVERLAY_ID = 'DYNAMIC_OVERLAY_ID'

CurvesCallback = Callable[[], Dict[Curve, utils.CurveDrawArgs]]

class PyCurveHandler:

    def __init__(self,
                 screen: pygame.Surface,
                 get_static_curves_callback: CurvesCallback,
                 get_dynamic_curves_callback: CurvesCallback,
                 vec_transformer: vec.VecTransformer = None,
                 ):
        self._parent_screen = screen
        self._get_static_curves_callback = get_static_curves_callback
        self._get_dynamic_curves_callback = get_dynamic_curves_callback
        self._vec_transformer = vec_transformer

        self.surface_manager = SurfaceManager(
            surface_draw_callbacks=[
                SurfaceRegistryArgs(CurveSurfaceType.STATIC_BUFFER_ID.value, self.redraw_static_buffer_curves),
                SurfaceRegistryArgs(CurveSurfaceType.DYNAMIC_BUFFER_ID.value, self.redraw_dynamic_buffer_curves),
                SurfaceRegistryArgs(CurveSurfaceType.STATIC_CURVES_ID.value, self.redraw_static_curves),
                SurfaceRegistryArgs(CurveSurfaceType.DYNAMIC_CURVES_ID.value, self.redraw_dynamic_curves),
                SurfaceRegistryArgs(CurveSurfaceType.STATIC_OVERLAY_ID.value, self.redraw_static_overlay),
                SurfaceRegistryArgs(CurveSurfaceType.DYNAMIC_OVERLAY_ID.value, self.redraw_dynamic_overlay),
            ]
        )

    def show_all(self):
        self.surface_manager.show_all()

    def hide_all(self):
        self.surface_manager.hide_all()

    def toggle_surface(self, curveSurfaceTypes: List[CurveSurfaceType]):
        self.surface_manager.toggle_visible([x.value for x in curveSurfaceTypes])

    def redraw_static_buffer_curves(self) -> pygame.Surface:
        curves = self._get_static_curves_callback()
        surf = help.init_surface(self._parent_screen.get_size())
        utils.draw_curves(
            curves={k: v.BufferArgs for k, v in curves.items()},
            surface=surf,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_dynamic_buffer_curves(self) -> pygame.Surface:
        curves = self._get_dynamic_curves_callback()
        surf = help.init_surface(self._parent_screen.get_size())
        utils.draw_curves(
            curves={k: v.BufferArgs for k, v in curves.items()},
            surface=surf,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_static_curves(self) -> pygame.Surface:
        curves = self._get_static_curves_callback()
        surf = help.init_surface(self._parent_screen.get_size())
        utils.draw_curves(
            curves={k: v.BaseArgs for k, v in curves.items()},
            surface=surf,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_dynamic_curves(self) -> pygame.Surface:
        curves = self._get_dynamic_curves_callback()
        surf = help.init_surface(self._parent_screen.get_size())
        utils.draw_curves(
            curves={k: v.BaseArgs for k, v in curves.items()},
            surface=surf,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_static_overlay(self) -> pygame.Surface:
        curves = self._get_static_curves_callback()
        surf = help.init_surface(self._parent_screen.get_size())
        utils.draw_curves(
            curves={k: v.OverlayArgs for k, v in curves.items()},
            surface=surf,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_dynamic_overlay(self) -> pygame.Surface:
        curves = self._get_dynamic_curves_callback()
        surf = help.init_surface(self._parent_screen.get_size())
        utils.draw_curves(
            curves={k: v.OverlayArgs for k, v in curves.items()},
            surface=surf,
            vec_transformer=self._vec_transformer
        )
        return surf

    def update(self):
        self.surface_manager.update_if_visible([
            CurveSurfaceType.DYNAMIC_BUFFER_ID.value,
            CurveSurfaceType.DYNAMIC_CURVES_ID.value,
            CurveSurfaceType.DYNAMIC_OVERLAY_ID.value
        ])

    def redraw(self):
        self.surface_manager.redraw([x.value for x in CurveSurfaceType])

    def invalidate(self):
        self.redraw()

    def render(self,
               surface: pygame.Surface):
        self.update()
        self.surface_manager.render(surface)


import pygame
from typing import Callable, Tuple, Dict
from cooptools.toggles import BooleanToggleable
from typing import List, Any, Callable
from dataclasses import dataclass, field
from cooptools.register import Register
import logging

logger = logging.getLogger(f"coopgame.surfacemanager")

DrawCallback = Callable[[], pygame.Surface]

@dataclass(frozen=True)
class SurfaceRegistryArgs:
    id: str
    callback: DrawCallback
    frame_update: bool = False

@dataclass
class SurfaceRegister:
    registry_args: SurfaceRegistryArgs
    visibility_toggle: BooleanToggleable = field(init=False, default_factory=lambda: BooleanToggleable(default=True))

class SurfaceManager:
    def __init__(self,
                 surface_draw_callbacks: List[SurfaceRegistryArgs] = None):
        self.surface_register: Register = Register[SurfaceRegister]()

        if surface_draw_callbacks:
            self.register_surface_ids(surface_draw_callbacks)

        self.surfaces: Dict[str, pygame.Surface] = {}

    @property
    def RegisteredSurfaceIds(self) -> List[str]:
        return list(self.surface_register.Registry.keys())

    @property
    def FrameUpdateSurfaceIds(self) -> List[str]:
        return [surf.registry_args.id for id, surf in self.surface_register.Registry.items() if surf.registry_args.frame_update]

    def register_surface_ids(self,
                             surface_draw_callbacks: List[SurfaceRegistryArgs]):
        self.surface_register.register(
            to_register=[SurfaceRegister(x) for x in surface_draw_callbacks],
            ids=[x.id for x in surface_draw_callbacks]
        )

    def invalidate(self, ids: List[str] = None):
        self.redraw(ids=ids)

    def redraw(self,
               ids: List[str] = None):
        if ids is None:
            ids = list(self.surface_register.Registry.keys())

        for id in ids:
            logger.debug(f"Drawing surface {id}...")
            self.surfaces[id] = self.surface_register.Registry[id].registry_args.callback()
            logger.debug(f"Surface {id} drawn")

    def update(self):
        self.update_if_visible(self.FrameUpdateSurfaceIds)

    def render(self,
               surface: pygame.Surface,
               at_pos: Tuple[int, int] = None,
               frame_update: bool = False):
        if frame_update:
            self.update()

        if at_pos is None:
            at_pos = (0, 0)

        for id, sr in self.surface_register.Registry.items():
            if sr.visibility_toggle.value:
                surface.blit(self.get_surfaces(ids=[id], dims=surface.get_size())[id], dest=at_pos)

    def get_surfaces(self,
                     ids: List[str],
                     dims: Tuple[int, int],
                     force_update: bool = False) -> Dict[str, pygame.Surface]:
        # update surfaces that dont exist or that dont match dims
        for id in ids:
            if id not in self.surfaces.keys() or \
                self.surfaces[id].get_size() != dims or \
                force_update:
                self.redraw([id])

        # return the surfaces
        return {id: self.surfaces[id] for id in ids}

    def get_toggled_state(self, ids: List[str]) -> Dict[str, bool]:
        return {
            x: self.surface_register.Registry[x].visibility_toggle.value for x in ids
        }

    def toggle_visible(self, ids: List[str]) -> Dict[str, bool]:
        # toggle
        [self.surface_register.Registry[x].visibility_toggle.toggle() for x in ids]

        return self.get_toggled_state(ids)

    def set_visiblility(self, ids: List[str], visible: bool):
        # toggle
        [self.surface_register.Registry[x].visibility_toggle.set_value(visible) for x in ids]
        return self.get_toggled_state(ids)

    def hide_all(self) -> Dict[str, bool]:
        self.set_visiblility(ids=self.RegisteredSurfaceIds, visible=False)
        return self.get_toggled_state(self.RegisteredSurfaceIds)

    def show_all(self) -> Dict[str, bool]:
        self.set_visiblility(ids=self.RegisteredSurfaceIds, visible=True)
        return self.get_toggled_state(self.RegisteredSurfaceIds)

    def update_if_visible(self, ids: List[str]):
        self.redraw([id for id, visible in self.get_toggled_state(ids).items() if visible])




from coopstructs.vectors import Vector3
from coopgame.models.primitives.coopquaternion import CoopQuaternion

class ObjectOrientationArgs:
    def __init__(self,
                 translation_vector: Vector3 = None,
                 rotation_point: Vector3 = None,
                 rotation_axis: Vector3 = None,
                 rotation_rads: float = None,
                 scale_vector: Vector3 = None,
                 scale_point: Vector3 = None,
                 ):
        self.rotation_axis = rotation_axis if rotation_axis else Vector3(0, 1, 0)
        self.rotation_point = rotation_point if rotation_point else Vector3(0, 0, 0)
        self.rotation_rads = rotation_rads if rotation_rads else 0
        self.scale_vector = scale_vector if scale_vector else Vector3(1, 1, 1)
        self.scale_point = scale_point if scale_point else Vector3(0, 0, 0)
        self.translation_vector = translation_vector if translation_vector else Vector3(0, 0, 0)
        self.quaternion = CoopQuaternion(axis=[0, 1, 0], angle=0)
        self._scale_adjustment = Vector3(0, 0, 0)

    def __str__(self):
        return f"Postion: {self.translation_vector}" \
               f"\nRotated: {self.quaternion.radians} around {self.quaternion.axis}" \
               f"\nScale: {self.scale_vector} at {self.scale_point}"

    def change_rotation(self,
                        rads_delta: float,
                        rotation_axis: Vector3 = None,
                        rotation_point: Vector3 = None):
        new_q = CoopQuaternion(axis=[rotation_axis.x, rotation_axis.y, rotation_axis.z], angle=rads_delta)

        self.quaternion = new_q * self.quaternion

    def change_translation(self, delta_vector: Vector3):
        self.translation_vector += delta_vector

    def change_scale(self, scale_vector: Vector3, scale_point: Vector3 = None):
        adj_scale_point = scale_point.hadamard_product(scale_vector) if scale_point else Vector3(0, 0, 0)
        self._scale_adjustment = self._scale_adjustment.hadamard_product(scale_vector) + scale_point - adj_scale_point
        self.scale_vector = self.scale_vector.hadamard_product(scale_vector)

    def reset_scale(self):
        self.scale_vector = Vector3(1, 1, 1)
        self.scale_point = Vector3(0, 0, 0)
        self._scale_adjustment = Vector3(0, 0, 0)

    @property
    def scale_adjustment(self):
        return self._scale_adjustment


import logging

from coopstructs.zones import ZoneManager, ZoneAlreadyExistsException, ZoneDoesntExistException, zoneLogger
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle
from cooptools.colors import Color
import pygame
from coopgame.pygamehelpers import draw_polygon, draw_text
from typing import List, Callable, Any
import coopgame.pygamehelpers as help
from coopstructs.toggles import IntegerRangeToggleable

class PyZoneHandler:

    def __init__(self):
        self.zm = ZoneManager()
        self.colors = {}
        self.active_zone_name: str = None
        self.layer_selector = IntegerRangeToggleable(min=0, max=0, starting_value=0)
        self.origin = None

    def set_active_zone(self, point: Vector2):
        zOptions = self.zm.member_zones([point])[point]
        self.active_zone_name = zOptions[self.layer_selector.value]

    def add_to_active_zone(self, point: Vector2, at_idx: int = None):
        if self.active_zone_name is not None:
            try:
                self.zm.add_to_zone(self.active_zone_name, point, at_idx=at_idx)
            except Exception as e:
                logging.error(f"Skip adding point {point}: {e}")

    def remove_from_active_zone(self):
        if self.active_zone_name is not None:
            self.zm.remove_last_point_from_zone(self.active_zone_name)

    def init_new_zone(self,
                      naming_provider: Callable[[], str] = None,
                      initial_points: List[Vector2] = None,
                      set_active: bool = True):
        name = naming_provider()

        if name is None:
            return

        color = Color.random()

        self.zm.init_new_zone(name, initial_points=initial_points)
        self.colors[name] = color
        if set_active: self.active_zone_name = name

    def draw_zones(self,
                   surface: pygame.Surface,
                   alpha: int = 90,
                   draw_zone_label: bool = True,
                   draw_point_labels: bool = True,
                   draw_scale_matrix=None,
                   excluded_zones: List[str] = None):
        for name, zone in self.zm.zones.items():
            if excluded_zones and name in excluded_zones:
                continue

            self.draw_zone(zone_name=name,
                           surface=surface,
                           alpha=alpha,
                           draw_zone_label=draw_zone_label,
                           draw_point_labels=draw_point_labels,
                           draw_scale_matrix=draw_scale_matrix)

    def draw_zone(self,
                  zone_name: str,
                  surface: pygame.Surface,
                  alpha: int = 90,
                  draw_zone_label: bool = True,
                  draw_point_labels: bool = True,
                  draw_scale_matrix=None,
                  ):

        zone = self.zm.zones[zone_name]
        if zone.valid:
            scaled_points = help.scaled_points(zone.boundary_points, transform_matrix=draw_scale_matrix)

            width = 0 if zone_name != self.active_zone_name else 10
            draw_polygon(surface, scaled_points, self.colors[zone_name], alpha=alpha, width=width)

            if draw_zone_label:
                center = help.scaled_points([zone.center], transform_matrix=draw_scale_matrix)[0]
                draw_text(zone_name, surface, offset_rect=Rectangle(center.x, center.y, 100, 100),
                          color=Color.furthest_color(self.colors[zone_name].value))

            if draw_point_labels:
                for idx, point in enumerate(zone.boundary_points):
                    draw_text(str(idx), surface, offset_rect=Rectangle(point.x, point.y, 100, 100),
                              color=Color.furthest_color(self.colors[zone_name].value))





from typing import List
from coopstructs.vectors import Vector2

class GameObject:

    def __init__(self,
                 name: str,
                 pos: Vector2,
                 tags: List[str] = None):
        self.tags = tags if tags else []
        self.name = name
        self.pos = pos

    def is_tagged(self, tags: List[str]):
        return set(tags).issubset(self.tags)




import pygame
import numpy as np
from typing import Tuple
from cooptools.colors import Color
import coopgame.pygamehelpers as help
from coopstructs.geometry.rectangle import Rectangle
from typing import List

"""
This implemenation is based off of the description on the pygame website docs at: https://pygame.readthedocs.io/en/latest/tiles/tiles.html
"""
class TileSet:
    def __init__(self, tiles: List[pygame.Surface] = None):
        self.tiles = tiles or []

class FileTileset(TileSet):
    def __init__(self, file, size=(32, 32), margin=1, spacing=1):
        super().__init__()
        self.file = file
        self.size = size
        self.margin = margin
        self.spacing = spacing
        self.image = pygame.image.load(file)
        self.rect = self.image.get_rect()
        self.load()

    def load(self):
        self.tiles = []
        x0 = y0 = self.margin
        w, h = self.rect.size
        dx = self.size[0] + self.spacing
        dy = self.size[1] + self.spacing

        for x in range(x0, w, dx):
            for y in range(y0, h, dy):
                tile = pygame.Surface(self.size)
                tile.blit(self.image, (0, 0), (x, y, *self.size))
                self.tiles.append(tile)

    def __str__(self):
        return f'{self.__class__.__name__} file:{self.file} tile:{self.size}'




class Tilemap:
    def __init__(self,
                 tileset: TileSet,
                 map_size=None,
                 render_on_init: bool = True,
                 surface_size: Tuple[int, int] = None,
                 buffer_tiles: Tuple[int, int] = None,
                 map: np.ndarray = None,
                 pxls: int = 32,
                 outline_tile_color: Color = None):

        self.pxls = pxls
        self._tile_size = (self.pxls, self.pxls)

        if map is not None:
            self.map = map
        elif map_size:
            self.map = np.zeros(map_size, dtype=int)
        else:
            self.map = np.zeros((10, 20), dtype=int)

        self.size = self.map.shape
        self.buffer_tiles = buffer_tiles

        self.tileset = tileset
        self.image = None

        self.render_surface_size = surface_size
        if buffer_tiles is not None:
            self.render_surface_size = (self.render_surface_size[0] + self.buffer_tiles[0] * self._tile_size[0],
                                        self.render_surface_size[1] + self.buffer_tiles[1] * self._tile_size[1])

        self.outline_tile_color = outline_tile_color
        if render_on_init:
            self.render(self.render_surface_size, outline_tiles_color=outline_tile_color)

    def render(self, surface_size: Tuple[int, int], outline_tiles_color: Color = None):
        h, w = self.map.shape
        rows, cols = self.map.shape

        # if self.buffer_tiles is not None:
        #     h -= self.buffer_tiles[1]
        #     w -= self.buffer_tiles[0]

        self.render_surface_size = surface_size
        self.image = pygame.Surface(surface_size)
        self.image.fill(Color.HOT_PINK.value)
        self.image.set_colorkey(Color.HOT_PINK.value)
        self.image.convert_alpha()

        tile_size_w = surface_size[0] / w
        tile_size_h = surface_size[1] / h
        self._tile_size = (tile_size_w, tile_size_h)

        for i in range(w):

            for j in range(h):
                tile = self.tileset.tiles[self.map[j, i]]
                scaled_tile = pygame.transform.scale(tile, (int(tile_size_w), int(tile_size_h)))
                self.image.blit(scaled_tile, (i*tile_size_w,j*tile_size_h))
                if outline_tiles_color is not None:
                    help.draw_box(self.image,
                                  Rectangle(x=i * tile_size_w,
                                            y=j * tile_size_h,
                                            width=tile_size_w,
                                            height=tile_size_h),
                                  color=outline_tiles_color,
                                  width=1)


    def set_value(self, value):
        self.map.fill(value)
        self.render(self.render_surface_size, outline_tiles_color=self.outline_tile_color)

    def set_random(self):
        n = len(self.tileset.tiles)
        self.map = np.random.randint(n, size=self.size)
        self.render(self.render_surface_size, outline_tiles_color=self.outline_tile_color)

    def __str__(self):
        return f'{self.__class__.__name__} {self.size}'

    @property
    def rect(self):
        return self.image.get_rect()

    @property
    def TileSize(self):
        return self._tile_size